import enum
import re
from abc import ABC
from typing import List, Optional, Union

from attr import Factory, attrs

from .utils import quote


class StringEnum(enum.Enum):
    def __str__(self):
        return self.name


class OnConflict(StringEnum):
    ROLLBACK = enum.auto()
    ABORT = enum.auto()
    FAIL = enum.auto()
    IGNORE = enum.auto()
    REPLACE = enum.auto()

    def __str__(self):
        return self.name


class OnDelete(StringEnum):
    SET_NULL = enum.auto()
    SET_DEFAULT = enum.auto()
    CASCADE = enum.auto()
    RESTRICT = enum.auto()
    NO_ACTION = enum.auto()

    def __str__(self):
        if self == OnDelete.SET_NULL:
            return "SET NULL"
        elif self == OnDelete.SET_DEFAULT:
            return "SET DEFAULT"
        elif self == OnDelete.NO_ACTION:
            return "NO ACTION"
        else:
            return super().__str__()


OnUpdate = OnDelete


class ForeignKeyMatch(StringEnum):
    SIMPLE = enum.auto()
    FULL = enum.auto()
    PARTIAL = enum.auto()


class CollatingSequence(StringEnum):
    BINARY = enum.auto()
    NOCASE = enum.auto()
    RTRIM = enum.auto()


class GeneratedColumnStorage(StringEnum):
    VIRTUAL = enum.auto()
    STORED = enum.auto()


class DefaultValue(StringEnum):
    CURRENT_TIME = enum.auto()
    CURRENT_TIMESTAMP = enum.auto()
    CURRENT_DATE = enum.auto()


class Node(ABC):
    def accept(self, visitor):
        """
        Accept a visitor implementing the visitor pattern.

        For example::

            class RenameVisitor:
                def __init__(self, old_name, new_name):
                    self.old_name = old_name
                    self.new_name = new_name

                def rename(self, node):
                    return node.accept(self)

                def visit_column(self, node):
                    if node.name == self.old_name:
                        return Column(self.new_name, node.definition)
                    else:
                        return node

                def visit_default(self, node):
                    return node

        The implementation uses introspection to avoid having to define ``accept``
        methods on every ``Node`` subclass and every ``visit_XYZ`` method on the visitor
        class.
        """
        name = snake_case(self.__class__.__name__)
        visit_method = getattr(visitor, "visit_" + name, None)
        if visit_method is not None:
            return visit_method(self)
        else:
            default_visit_method = getattr(visitor, "visit_default", None)
            if default_visit_method is not None:
                return default_visit_method(self)
            else:
                return None

    def as_string(self, *, p: bool) -> str:
        raise NotImplementedError

    def __str__(self) -> str:
        return self.as_string(p=True)


class BaseConstraint(Node):
    pass


class Expression(Node):
    pass


@attrs(auto_attribs=True)
class CreateTableStatement(Node):
    """
    A SQL ``CREATE TABLE`` statement.
    """

    name: Union[str, "TableName"]
    columns: List["Column"]
    constraints: List[BaseConstraint] = Factory(list)
    as_select: Optional[Node] = None
    temporary: bool = False
    without_rowid: bool = False
    if_not_exists: bool = False

    def as_string(self, *, p: bool) -> str:
        builder = ["CREATE "]
        if self.temporary:
            builder.append("TEMPORARY ")
        builder.append("TABLE ")
        if self.if_not_exists:
            builder.append("IF NOT EXISTS ")

        if isinstance(self.name, TableName):
            builder.append(str(self.name))
        else:
            builder.append(quote(self.name))

        if self.as_select:
            builder.append(" AS ")
            builder.append(str(self.as_select))
            return "".join(builder)

        builder.append("(\n")
        for i, column in enumerate(self.columns):
            builder.append(str(column))
            if i < len(self.columns) - 1 or self.constraints:
                builder.append(",")
            builder.append("\n")

        for i, constraint in enumerate(self.constraints):
            builder.append(str(constraint))
            if i < len(self.constraints) - 1:
                builder.append(",")
            builder.append("\n")

        builder.append(")")
        if self.without_rowid:
            builder.append(" WITHOUT ROWID")

        return "".join(builder)


@attrs(auto_attribs=True)
class SelectStatement(Node):
    """
    A SQL ``SELECT`` statement.
    """

    columns: List[Expression]

    def as_string(self, *, p: bool) -> str:
        raise NotImplementedError


@attrs(auto_attribs=True)
class Column(Node):
    name: str
    definition: Optional["ColumnDefinition"]

    def as_string(self, *, p: bool) -> str:
        if self.definition is None:
            return quote(self.name)
        else:
            definition = self.definition.as_string(p=False)
            return f"{quote(self.name)} {definition}"


@attrs(auto_attribs=True)
class ColumnType(Node):
    name: str
    args: List[int]

    def as_string(self, *, p: bool) -> str:
        joined_args = ", ".join(map(str, self.args))
        return f"{quote(self.name)}({joined_args})"


@attrs(auto_attribs=True)
class ColumnDefinition(Node):
    type: Optional[Union[ColumnType, str]] = None
    default: Optional[Expression] = None
    constraints: List[BaseConstraint] = Factory(list)

    def as_string(self, *, p: bool) -> str:
        builder = []
        if self.type is not None:
            builder.append(" ")
            builder.append(str(self.type))

        for constraint in self.constraints:
            builder.append(" ")
            builder.append(str(constraint))

        if self.default is not None:
            builder.append(" DEFAULT ")
            builder.append(str(self.default))

        return "".join(builder)


@attrs(auto_attribs=True)
class CheckConstraint(BaseConstraint):
    expr: Expression

    def as_string(self, *, p: bool) -> str:
        e = self.expr.as_string(p=False)
        return f"CHECK({e})"


@attrs(auto_attribs=True)
class NamedConstraint(BaseConstraint):
    name: str
    constraint: BaseConstraint

    def as_string(self, *, p: bool) -> str:
        return f"CONSTRAINT {quote(self.name)} {self.constraint}"


@attrs(auto_attribs=True)
class NotNullConstraint(BaseConstraint):
    on_conflict: Optional[OnConflict] = None

    def as_string(self, *, p: bool) -> str:
        if self.on_conflict is not None:
            return f"NOT NULL {self.on_conflict}"
        else:
            return "NOT NULL"


@attrs(auto_attribs=True)
class PrimaryKeyConstraint(BaseConstraint):
    ascending: Optional[bool] = None
    on_conflict: Optional[OnConflict] = None
    autoincrement: bool = False

    def as_string(self, *, p: bool) -> str:
        builder = ["PRIMARY KEY"]
        if self.ascending is not None:
            if self.ascending:
                builder.append(" ASC")
            else:
                builder.append(" DESC")

        if self.on_conflict is not None:
            builder.append(" ")
            builder.append(str(self.on_conflict))

        if self.autoincrement:
            builder.append(" AUTOINCREMENT")

        return "".join(builder)


@attrs(auto_attribs=True)
class CollateConstraint(BaseConstraint):
    sequence: CollatingSequence

    def as_string(self, *, p: bool) -> str:
        return f"COLLATE {self.sequence}"


@attrs(auto_attribs=True)
class ForeignKeyConstraint(BaseConstraint):
    columns: List[str]
    foreign_table: str
    foreign_columns: List[str]
    on_delete: Optional[OnDelete] = None
    on_update: Optional[OnUpdate] = None
    match: Optional[ForeignKeyMatch] = None
    deferrable: Optional[bool] = None
    initially_deferred: Optional[bool] = None

    def as_string(self, *, p: bool) -> str:
        builder = []
        if self.columns:
            builder.append("FOREIGN KEY (")
            builder.append(", ".join(map(str, self.columns)))
            builder.append(" ")

        builder.append("REFERENCES ")
        builder.append(self.foreign_table)
        if self.foreign_columns:
            builder.append("(")
            builder.append(", ".join(map(str, self.foreign_columns)))
            builder.append(")")

        if self.on_delete:
            builder.append(f" ON DELETE {self.on_delete}")

        if self.on_update:
            builder.append(f" ON UPDATE {self.on_update}")

        if self.match:
            builder.append(f" MATCH {self.match}")

        if self.deferrable is not None:
            builder.append(" ")
            if not self.deferrable:
                builder.append("NOT")
            builder.append(" DEFERRABLE")

            if self.initially_deferred is not None:
                builder.append(" INITIALLY ")
                builder.append("DEFERRED" if self.initially_deferred else "IMMEDIATE")

        return "".join(builder)


@attrs(auto_attribs=True)
class UniqueConstraint(BaseConstraint):
    on_conflict: Optional[OnConflict] = None

    def as_string(self, *, p: bool) -> str:
        if self.on_conflict is not None:
            return f"UNIQUE {self.on_conflict}"
        else:
            return "UNIQUE"


@attrs(auto_attribs=True)
class GeneratedColumnConstraint(BaseConstraint):
    expression: Expression
    storage: Optional[GeneratedColumnStorage] = None

    def as_string(self, *, p: bool) -> str:
        e = self.expression.as_string(p=False)
        storage_string = "" if self.storage is None else " " + str(self.storage)
        return f"GENERATED ALWAYS AS ({e}){storage_string}"


@attrs(auto_attribs=True)
class Infix(Expression):
    operator: str
    left: Expression
    right: Expression

    def as_string(self, *, p: bool) -> str:
        left = self.left.as_string(p=True)
        right = self.right.as_string(p=True)
        core = f"{left} {self.operator} {right}"
        return f"({core})" if p else core


@attrs(auto_attribs=True)
class ExpressionList(Expression):
    values: List[Expression]

    def as_string(self, *, p: bool) -> str:
        return "(" + ", ".join(v.as_string(p=False) for v in self.values) + ")"


@attrs(auto_attribs=True)
class Identifier(Expression):
    value: str

    def as_string(self, *, p: bool) -> str:
        return quote(self.value)


@attrs(auto_attribs=True)
class String(Expression):
    value: str

    def as_string(self, *, p: bool) -> str:
        escaped = self.value.replace("'", "''")
        return f"'{escaped}'"


@attrs(auto_attribs=True)
class Blob(Expression):
    value: bytes

    def as_string(self, *, p: bool) -> str:
        return "X'" + "".join(f"{hex(b)[2:]:0>2}" for b in self.value) + "'"


@attrs(auto_attribs=True)
class Integer(Expression):
    value: int

    def as_string(self, *, p: bool) -> str:
        return str(self.value)


@attrs(auto_attribs=True)
class Null(Expression):
    def as_string(self, *, p: bool) -> str:
        return "NULL"


@attrs(auto_attribs=True)
class Boolean(Expression):
    value: bool

    def as_string(self, *, p: bool) -> str:
        return "TRUE" if self.value else "FALSE"


@attrs(auto_attribs=True)
class TableName(Node):
    schema_name: str
    table_name: str

    def as_string(self, *, p: bool) -> str:
        return f"{quote(self.schema_name)}.{quote(self.table_name)}"


def snake_case(s: str) -> str:
    def snake_case_replacer(match: re.Match) -> str:
        text = match.group(0)
        return text[0] + "_" + text[1]

    name = re.sub(r"[a-z][A-Z]", snake_case_replacer, s)
    return name.lower()



import enum
import string
from typing import Any, List, Optional, Tuple, Union

from attr import attrs

from .exceptions import SQLiteParserError


class TokenType(enum.Enum):
    KEYWORD = enum.auto()
    IDENTIFIER = enum.auto()
    LEFT_PARENTHESIS = enum.auto()
    RIGHT_PARENTHESIS = enum.auto()
    COMMA = enum.auto()
    SEMICOLON = enum.auto()
    DOT = enum.auto()
    NOT_EQ = enum.auto()
    GREATER_THAN = enum.auto()
    LESS_THAN = enum.auto()
    GREATER_THAN_OR_EQ = enum.auto()
    LESS_THAN_OR_EQ = enum.auto()
    STRING = enum.auto()
    BLOB = enum.auto()
    INTEGER = enum.auto()
    EOF = enum.auto()
    UNKNOWN = enum.auto()


@attrs(auto_attribs=True)
class Token:
    type: TokenType
    value: Any
    line: int
    column: int


CheckTokenType = List[Union[str, TokenType, Tuple[TokenType, str]]]


class Lexer:
    program: str
    index: int
    line: int
    column: int

    pushed_token: Optional[Token]
    current_token: Token

    def __init__(self, program: str) -> None:
        self.program = program
        self.index = 0
        self.line = 1
        self.column = 1

        self.pushed_token = None
        self.advance()

    def current(self) -> Token:
        return (
            self.pushed_token if self.pushed_token is not None else self.current_token
        )

    def check(self, types_and_values: CheckTokenType) -> Token:
        """
        Checks that the current token matches at least one of the items in
        ``types_and_values``. If the current token does match, it is returned. If not, a
        SQLiteParserError is raised.

        Each item in ``types_and_values`` must be one of the following types:

          - A string, which matches a ``KEYWORD`` token with that string's value
          - A ``TokenType`` instance, which matches a token with that type
          - A ``(TokenType, string)`` pair, which matches a token with that type and
              that value.
        """
        token = self.current()
        if token is None:
            raise SQLiteParserError("premature end of input")

        if token.type == TokenType.UNKNOWN:
            raise SQLiteParserError(f"unknown token: {token.value!r}")

        for type_or_value in types_and_values:
            if isinstance(type_or_value, str):
                if token.type == TokenType.KEYWORD and token.value == type_or_value:
                    return token
            elif isinstance(type_or_value, tuple):
                if token.type == type_or_value[0] and token.value == type_or_value[1]:
                    return token
            else:
                if token.type == type_or_value:
                    return token

        expected = " or ".join(
            type_or_value if isinstance(type_or_value, str) else repr(type_or_value)
            for type_or_value in types_and_values
        )
        raise SQLiteParserError(f"expected {expected}, got {token.value!r}")

    def advance(self, expecting: Optional[CheckTokenType] = None) -> Token:
        if self.pushed_token is not None:
            ret = self.pushed_token
            self.pushed_token = None
            self.current_token = ret
            return ret

        self.read_whitespace()

        if self.done():
            if expecting is not None:
                raise SQLiteParserError("premature end of input")

            self.current_token = Token(TokenType.EOF, "", self.line, self.column)
            return self.current_token

        self.current_token = self._advance()
        if expecting is not None:
            self.check(expecting)

        self.read_whitespace()
        return self.current_token

    def _advance(self) -> Token:
        c = self.c()
        if c.upper() == "X" and self.peek() == "'":
            return self.read_blob()
        elif c.isalpha():
            return self.read_symbol()
        elif c.isdigit():
            return self.read_integer()
        elif c == "'":
            return self.read_generic_string(TokenType.STRING, "'")
        elif c == '"':
            return self.read_generic_string(TokenType.IDENTIFIER, '"')
        elif c == "`":
            return self.read_generic_string(TokenType.IDENTIFIER, "`")
        elif c == "[":
            return self.read_generic_string(
                TokenType.IDENTIFIER, "]", allow_doubling=False
            )
        elif c == "(":
            return self.character_token(TokenType.LEFT_PARENTHESIS)
        elif c == ")":
            return self.character_token(TokenType.RIGHT_PARENTHESIS)
        elif c == ",":
            return self.character_token(TokenType.COMMA)
        elif c == ";":
            return self.character_token(TokenType.SEMICOLON)
        elif c == ".":
            return self.character_token(TokenType.DOT)
        elif self.prefix(2) == "!=":
            return self.multi_character_token(TokenType.NOT_EQ, 2)
        elif self.prefix(2) == ">=":
            return self.multi_character_token(TokenType.GREATER_THAN_OR_EQ, 2)
        elif self.prefix(2) == "<=":
            return self.multi_character_token(TokenType.LESS_THAN_OR_EQ, 2)
        elif c == ">":
            return self.character_token(TokenType.GREATER_THAN)
        elif c == "<":
            return self.character_token(TokenType.GREATER_THAN)
        else:
            return self.character_token(TokenType.UNKNOWN)

    def push(self, token: Token) -> None:
        if self.pushed_token is not None:
            raise SQLiteParserError("token already pushed")

        self.pushed_token = token

    def done(self) -> bool:
        return self.index == len(self.program)

    def read_whitespace(self) -> None:
        while not self.done() and self.c().isspace():
            self.next_character()

    def read_symbol(self) -> Token:
        start = self.index
        start_column = self.column
        while not self.done() and is_symbol_character(self.c()):
            self.next_character()

        value = self.program[start : self.index]
        if value.upper() in SQL_KEYWORDS:
            return Token(
                type=TokenType.KEYWORD,
                value=value.upper(),
                line=self.line,
                column=start_column,
            )
        else:
            return Token(
                type=TokenType.IDENTIFIER,
                value=value,
                line=self.line,
                column=start_column,
            )

    def read_integer(self) -> Token:
        start = self.index
        start_column = self.column
        while not self.done() and self.c().isdigit():
            self.next_character()

        value = self.program[start : self.index]
        return Token(
            type=TokenType.INTEGER, value=value, line=self.line, column=start_column
        )

    def read_blob(self) -> Token:
        start_column = self.column
        characters = []

        self.next_character()
        self.next_character()
        while not self.done() and self.c() != "'":
            d1 = self.c()

            self.next_character()
            if self.done():
                raise SQLiteParserError("unterminated blob literal")

            d2 = self.c()

            if d1 not in string.hexdigits:
                raise SQLiteParserError(f"invalid hex digit: {d1!r}")

            if d2 not in string.hexdigits:
                raise SQLiteParserError(f"invalid hex digit: {d2!r}")

            characters.append(int(d1 + d2, base=16))
            self.next_character()

        if self.done():
            raise SQLiteParserError("unterminated blob literal")
        else:
            self.next_character()

        return Token(
            type=TokenType.BLOB,
            value=bytes(characters),
            line=self.line,
            column=start_column,
        )

    def read_generic_string(
        self, token_type: TokenType, delimiter: str, *, allow_doubling: bool = True
    ) -> Token:
        start_column = self.column
        characters = []

        self.next_character()
        while not self.done():
            if self.c() == delimiter:
                # SQL escapes quotes by doubling them.
                if allow_doubling and self.prefix(2) == delimiter * 2:
                    characters.append(delimiter)
                    self.next_character()
                    self.next_character()
                else:
                    break
            else:
                characters.append(self.c())
                self.next_character()

        if self.done():
            raise SQLiteParserError(f"expected delimiter: {delimiter}")
        else:
            self.next_character()

        return Token(
            type=token_type,
            value="".join(characters),
            line=self.line,
            column=start_column,
        )

    def next_character(self) -> None:
        if self.c() == "\n":
            self.line += 1
            self.column = 1
        else:
            self.column += 1

        self.index += 1

    def c(self) -> str:
        return self.prefix(1)

    def peek(self, n: int = 1) -> Optional[str]:
        if self.index + n >= len(self.program):
            return None

        return self.program[self.index + n]

    def prefix(self, length: int) -> str:
        return self.program[self.index : self.index + length]

    def character_token(self, type: TokenType) -> Token:
        return self.multi_character_token(type, 1)

    def multi_character_token(self, type: TokenType, length: int) -> Token:
        value = self.program[self.index : self.index + length]
        line = self.line
        column = self.column
        for _ in range(length):
            self.next_character()
        return Token(type=type, value=value, line=line, column=column)


def is_symbol_character(c: str) -> bool:
    return c.isalpha() or c.isdigit() or c == "_"


# According to https://sqlite.org/lang_keywords.html
SQL_KEYWORDS = {
    "ABORT",
    "ACTION",
    "ADD",
    "AFTER",
    "ALL",
    "ALTER",
    "ALWAYS",
    "ANALYZE",
    "AND",
    "AS",
    "ASC",
    "ATTACH",
    "AUTOINCREMENT",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BY",
    "CASCADE",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "CONFLICT",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "DATABASE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DELETE",
    "DESC",
    "DETACH",
    "DISTINCT",
    "DO",
    "DROP",
    "EACH",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXCLUDE",
    "EXCLUSIVE",
    "EXISTS",
    "EXPLAIN",
    "FAIL",
    "FILTER",
    "FIRST",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "GENERATED",
    "GLOB",
    "GROUP",
    "GROUPS",
    "HAVING",
    "IF",
    "IGNORE",
    "IMMEDIATE",
    "IN",
    "INDEX",
    "INDEXED",
    "INITIALLY",
    "INNER",
    "INSERT",
    "INSTEAD",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "KEY",
    "LAST",
    "LEFT",
    "LIKE",
    "LIMIT",
    "MATCH",
    "MATERIALIZED",
    "NATURAL",
    "NO",
    "NOT",
    "NOTHING",
    "NOTNULL",
    "NULL",
    "NULLS",
    "OF",
    "OFFSET",
    "ON",
    "OR",
    "ORDER",
    "OTHERS",
    "OUTER",
    "OVER",
    "PARTITION",
    "PLAN",
    "PRAGMA",
    "PRECEDING",
    "PRIMARY",
    "QUERY",
    "RAISE",
    "RANGE",
    "RECURSIVE",
    "REFERENCES",
    "REGEXP",
    "REINDEX",
    "RELEASE",
    "RENAME",
    "REPLACE",
    "RESTRICT",
    "RETURNING",
    "RIGHT",
    "ROLLBACK",
    "ROW",
    "ROWS",
    "SAVEPOINT",
    "SELECT",
    "SET",
    "TABLE",
    "TEMP",
    "TEMPORARY",
    "THEN",
    "TIES",
    "TO",
    "TRANSACTION",
    "TRIGGER",
    "UNBOUNDED",
    "UNION",
    "UNIQUE",
    "UPDATE",
    "USING",
    "VACUUM",
    "VALUES",
    "VIEW",
    "VIRTUAL",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
    "WITHOUT",
}



from typing import List, Optional, Union

from . import ast
from .exceptions import SQLiteParserError, SQLiteParserImpossibleError
from .lexer import Lexer, TokenType


def parse(program: str, *, debug: bool = False) -> List[ast.Node]:
    """
    Parse the SQL program into a list of AST objects.
    """
    lexer = Lexer(program)
    parser = Parser(lexer, debug=debug)
    return parser.parse()


def parse_column(column_string: str, *, debug: bool = False) -> ast.Column:
    """
    Parse a single column from a ``CREATE TABLE`` statement.
    """
    lexer = Lexer(column_string)
    parser = Parser(lexer, debug=debug)
    return parser.parse_column()


def debuggable(f):
    def wrapped(self: "Parser", *args, **kwargs):
        name = f.__name__
        if self.debug:
            indent = "  " * (self.debug_indent * 2)
            self.debug_indent += 1
            print(indent + f"{name} (token = {self.lexer.current().value!r})")

        r = f(self, *args, **kwargs)

        if self.debug:
            print(
                indent
                + f"{name} returned (token = {self.lexer.current().value!r}, "
                + f"value = {r!r})"
            )
            self.debug_indent -= 1

        return r

    return wrapped


class Parser:
    """
    The SQL parser.

    It is implemented as a recursive-descent parser. Each match_XYZ method obeys the
    following protocol:

      - It assumes that the lexer is positioned at the first token of the fragment to be
        matched, e.g. match_create_statement assumes that self.lexer.current() will
        return the CREATE token.

      - It leaves the lexer positioned at one past the last token of the fragment.
    """

    lexer: Lexer
    debug: bool
    debug_indent: int

    def __init__(self, lexer: Lexer, *, debug: bool = False) -> None:
        self.lexer = lexer
        self.debug = debug
        self.debug_indent = 0

    def parse(self) -> List[ast.Node]:
        statements = []
        while True:
            if self.lexer.done():
                break

            statement = self.match_statement()
            statements.append(statement)

            if not self.lexer.done():
                self.lexer.advance(expecting=[TokenType.SEMICOLON])

        return statements

    def parse_column(self) -> ast.Column:
        column = self.match_column()
        if not self.lexer.done():
            raise SQLiteParserError("trailing input")
        return column

    @debuggable
    def match_statement(self) -> ast.Node:
        token = self.lexer.current()
        if token.type == TokenType.KEYWORD:
            if token.value == "CREATE":
                return self.match_create_statement()
            elif token.value == "SELECT":
                return self.match_select_statement()
            else:
                raise SQLiteParserError(f"unexpected keyword: {token.value}")
        else:
            raise SQLiteParserError(f"unexpected token type: {token.type}")

    @debuggable
    def match_create_statement(self) -> ast.CreateTableStatement:
        token = self.lexer.advance(expecting=["TABLE", "TEMPORARY", "TEMP"])
        if token.value in ("TEMPORARY", "TEMP"):
            temporary = True
            self.lexer.advance(expecting=["TABLE"])
        else:
            temporary = False

        token = self.lexer.advance(expecting=["IF", TokenType.IDENTIFIER, "TEMP"])
        if token.value == "IF":
            self.lexer.advance(expecting=["NOT"])
            self.lexer.advance(expecting=["EXISTS"])
            if_not_exists = True
            name_token = self.lexer.advance(expecting=[TokenType.IDENTIFIER])
        else:
            if_not_exists = False
            name_token = token

        token = self.lexer.advance(
            expecting=[TokenType.DOT, TokenType.LEFT_PARENTHESIS]
        )
        if token.type == TokenType.DOT:
            table_name_token = self.lexer.advance(expecting=[TokenType.IDENTIFIER])
            name = ast.TableName(name_token.value, table_name_token.value)
            self.lexer.advance(expecting=[TokenType.LEFT_PARENTHESIS])
        else:
            name = name_token.value

        columns = []
        constraints: List[ast.BaseConstraint] = []
        while True:
            token = self.lexer.advance()
            if token.type == TokenType.RIGHT_PARENTHESIS:
                break

            column_or_constraint = self.match_column_or_constraint()
            if isinstance(column_or_constraint, ast.Column):
                if constraints:
                    raise SQLiteParserError

                columns.append(column_or_constraint)
            else:
                constraints.append(column_or_constraint)

            token = self.lexer.check([TokenType.COMMA, TokenType.RIGHT_PARENTHESIS])
            if token.type == TokenType.RIGHT_PARENTHESIS:
                break

        token = self.lexer.advance()
        if token is not None:
            if token.type == TokenType.KEYWORD and token.value == "WITHOUT":
                self.lexer.advance(expecting=[(TokenType.IDENTIFIER, "ROWID")])
                without_rowid = True
            else:
                self.lexer.push(token)
                without_rowid = False
        else:
            without_rowid = False

        return ast.CreateTableStatement(
            name=name,
            columns=columns,
            constraints=constraints,
            as_select=None,
            temporary=temporary,
            without_rowid=without_rowid,
            if_not_exists=if_not_exists,
        )

    @debuggable
    def match_select_statement(self) -> ast.SelectStatement:
        self.lexer.advance()
        e = self.match_expression()
        return ast.SelectStatement(columns=[e])

    @debuggable
    def match_column_or_constraint(self) -> Union[ast.BaseConstraint, ast.Column]:
        token = self.lexer.current()
        if token.type == TokenType.KEYWORD and token.value == "FOREIGN":
            return self.match_foreign_key_constraint()
        elif token.type == TokenType.IDENTIFIER:
            return self.match_column()
        else:
            raise SQLiteParserError("expected start of column or constraint")

    @debuggable
    def match_column(self) -> ast.Column:
        name_token = self.lexer.check([TokenType.IDENTIFIER])

        column_type = self.match_column_type()
        if column_type is None:
            return ast.Column(name=name_token.value, definition=None)

        token = self.lexer.current()
        constraints = []
        default = None
        while True:
            if token is not None and token.type == TokenType.KEYWORD:
                if token.value == "PRIMARY":
                    constraints.append(self.match_primary_key_constraint())
                elif token.value == "NOT":
                    constraints.append(self.match_not_null_constraint())
                elif token.value == "CHECK":
                    constraints.append(self.match_check_constraint())
                elif token.value == "COLLATE":
                    constraints.append(self.match_collate_constraint())
                elif token.value == "REFERENCES":
                    constraints.append(self.match_foreign_key_clause(columns=[]))
                elif token.value == "UNIQUE":
                    constraints.append(self.match_unique_constraint())
                elif token.value == "GENERATED" or token.value == "AS":
                    constraints.append(self.match_generated_column_constraint())
                elif token.value == "DEFAULT":
                    default = self.match_default_clause()
                elif token.value == "NULL":
                    # Django's ORM will produce columns with a NULL constraint, and
                    # SQLite accepts it although it is not documented and appears to
                    # have no effect.
                    self.lexer.advance()
                else:
                    raise SQLiteParserError(f"unexpected keyword: {token.value}")
            else:
                break

            token = self.lexer.current()

        return ast.Column(
            name=name_token.value,
            definition=ast.ColumnDefinition(
                type=column_type,
                default=default,
                constraints=constraints,
            ),
        )

    @debuggable
    def match_column_type(self) -> Optional[Union[str, ast.ColumnType]]:
        type_token = self.lexer.advance()
        if type_token.type != TokenType.IDENTIFIER:
            return None

        token = self.lexer.advance()
        if token.type == TokenType.LEFT_PARENTHESIS:
            args = []
            while True:
                token = self.lexer.advance()
                if token.type == TokenType.RIGHT_PARENTHESIS:
                    break
                elif token.type == TokenType.INTEGER:
                    args.append(int(token.value))
                    token = self.lexer.advance()
                    if token.type == TokenType.COMMA:
                        continue
                    elif token.type == TokenType.RIGHT_PARENTHESIS:
                        break
                    else:
                        raise SQLiteParserError("expected comma or right parenthesis")

            token = self.lexer.advance()
            return ast.ColumnType(name=type_token.value, args=args)
        else:
            # SQL allows for multi-word column types, e.g. `smallint unsigned`.
            name_parts = [type_token.value]
            while token.type == TokenType.IDENTIFIER:
                name_parts.append(token.value)
                token = self.lexer.advance()
            return " ".join(name_parts)

    @debuggable
    def match_foreign_key_constraint(self) -> ast.ForeignKeyConstraint:
        self.lexer.check(["FOREIGN"])
        self.lexer.advance(expecting=["KEY"])

        self.lexer.advance(expecting=[TokenType.LEFT_PARENTHESIS])
        self.lexer.advance()
        columns = self.match_identifier_list()
        self.lexer.check([TokenType.RIGHT_PARENTHESIS])
        self.lexer.advance()
        return self.match_foreign_key_clause(columns=columns)

    @debuggable
    def match_foreign_key_clause(
        self, *, columns: List[str]
    ) -> ast.ForeignKeyConstraint:
        self.lexer.check(["REFERENCES"])

        foreign_table = self.lexer.advance(expecting=[TokenType.IDENTIFIER]).value

        token = self.lexer.advance()
        if token is not None and token.type == TokenType.LEFT_PARENTHESIS:
            self.lexer.advance()
            foreign_columns = self.match_identifier_list()
            self.lexer.check([TokenType.RIGHT_PARENTHESIS])
            token = self.lexer.advance()
        else:
            foreign_columns = []

        on_delete = None
        on_update = None
        match = None
        deferrable = None
        initially_deferred = None

        while True:
            if token is None:
                break

            if token.value == "ON":
                delete_or_update = self.lexer.advance(expecting=["DELETE", "UPDATE"])
                token = self.lexer.advance(
                    expecting=["SET", "CASCADE", "RESTRICT", "NO"]
                )
                if token.value == "SET":
                    token = self.lexer.advance(expecting=["NULL", "DEFAULT"])
                    value = (
                        ast.OnDelete.SET_NULL
                        if token.value == "NULL"
                        else ast.OnDelete.SET_DEFAULT
                    )
                elif token.value == "NO":
                    self.lexer.advance(expecting=["ACTION"])
                    value = ast.OnDelete.NO_ACTION
                elif token.value == "CASCADE":
                    value = ast.OnDelete.CASCADE
                elif token.value == "RESTRICT":
                    value = ast.OnDelete.RESTRICT
                else:
                    raise SQLiteParserImpossibleError(token.value)

                if delete_or_update.value == "DELETE":
                    on_delete = value
                else:
                    on_update = value
            elif token.value == "MATCH":
                match_token = self.lexer.advance(
                    expecting=["SIMPLE", "FULL", "PARTIAL"]
                )
                if match_token.value == "FULL":
                    match = ast.ForeignKeyMatch.FULL
                elif match_token.value == "PARTIAL":
                    match = ast.ForeignKeyMatch.PARTIAL
                else:
                    match = ast.ForeignKeyMatch.SIMPLE
            elif token.value == "NOT" or token.value == "DEFERRABLE":
                if token.value == "NOT":
                    self.lexer.advance(expecting=["DEFERRABLE"])
                    deferrable = False
                else:
                    deferrable = True

                token = self.lexer.advance()
                if not (token.type == TokenType.KEYWORD and token.value == "INITIALLY"):
                    break

                token = self.lexer.advance(expecting=["DEFERRED", "IMMEDIATE"])
                initially_deferred = bool(token.value == "DEFERRED")
                self.lexer.advance()
                break
            else:
                break

            token = self.lexer.advance()

        return ast.ForeignKeyConstraint(
            columns=columns,
            foreign_table=foreign_table,
            foreign_columns=foreign_columns,
            on_delete=on_delete,
            on_update=on_update,
            match=match,
            deferrable=deferrable,
            initially_deferred=initially_deferred,
        )

    @debuggable
    def match_not_null_constraint(self) -> ast.NotNullConstraint:
        self.lexer.check(["NOT"])
        self.lexer.advance(expecting=["NULL"])
        token = self.lexer.advance()
        if (
            token is not None
            and token.type == TokenType.KEYWORD
            and token.value == "ON"
        ):
            on_conflict = self.match_on_conflict_clause()
        else:
            on_conflict = None
        return ast.NotNullConstraint(on_conflict=on_conflict)

    @debuggable
    def match_primary_key_constraint(self) -> ast.PrimaryKeyConstraint:
        self.lexer.check(["PRIMARY"])
        self.lexer.advance(expecting=["KEY"])
        token = self.lexer.advance()

        if token is None or token.type != TokenType.KEYWORD:
            return ast.PrimaryKeyConstraint()

        ascending: Optional[bool]
        if token.value == "ASC":
            ascending = True
            token = self.lexer.advance()
        elif token.value == "DESC":
            ascending = False
            token = self.lexer.advance()
        else:
            ascending = None

        if token.type != TokenType.KEYWORD:
            return ast.PrimaryKeyConstraint(ascending=ascending)

        if token.value == "ON":
            on_conflict = self.match_on_conflict_clause()
            token = self.lexer.current()
        else:
            on_conflict = None

        if token.type != TokenType.KEYWORD:
            return ast.PrimaryKeyConstraint(
                ascending=ascending, on_conflict=on_conflict
            )

        if token.value == "AUTOINCREMENT":
            autoincrement = True
            self.lexer.advance()
        else:
            autoincrement = False

        return ast.PrimaryKeyConstraint(
            ascending=ascending, on_conflict=on_conflict, autoincrement=autoincrement
        )

    @debuggable
    def match_check_constraint(self) -> ast.CheckConstraint:
        self.lexer.check(["CHECK"])
        self.lexer.advance(expecting=[TokenType.LEFT_PARENTHESIS])
        self.lexer.advance()
        expr = self.match_expression()
        self.lexer.check([TokenType.RIGHT_PARENTHESIS])
        self.lexer.advance()
        return ast.CheckConstraint(expr)

    @debuggable
    def match_collate_constraint(self) -> ast.CollateConstraint:
        self.lexer.check(["COLLATE"])
        sequence = self.lexer.advance(
            expecting=[
                (TokenType.IDENTIFIER, "BINARY"),
                (TokenType.IDENTIFIER, "NOCASE"),
                (TokenType.IDENTIFIER, "RTRIM"),
            ]
        ).value
        self.lexer.advance()
        if sequence == "NOCASE":
            return ast.CollateConstraint(ast.CollatingSequence.NOCASE)
        elif sequence == "RTRIM":
            return ast.CollateConstraint(ast.CollatingSequence.RTRIM)
        elif sequence == "BINARY":
            return ast.CollateConstraint(ast.CollatingSequence.BINARY)
        else:
            raise SQLiteParserImpossibleError(sequence)

    @debuggable
    def match_unique_constraint(self) -> ast.UniqueConstraint:
        self.lexer.check(["UNIQUE"])
        token = self.lexer.advance()
        if (
            token is not None
            and token.type == TokenType.KEYWORD
            and token.value == "ON"
        ):
            on_conflict = self.match_on_conflict_clause()
        else:
            on_conflict = None
        return ast.UniqueConstraint(on_conflict=on_conflict)

    @debuggable
    def match_generated_column_constraint(self) -> ast.GeneratedColumnConstraint:
        token = self.lexer.check(["GENERATED", "AS"])
        if token.value == "GENERATED":
            self.lexer.advance(expecting=["ALWAYS"])
            self.lexer.advance(expecting=["AS"])

        self.lexer.advance(expecting=[TokenType.LEFT_PARENTHESIS])
        self.lexer.advance()
        e = self.match_expression()
        self.lexer.check([TokenType.RIGHT_PARENTHESIS])

        token = self.lexer.advance()
        storage: Optional[ast.GeneratedColumnStorage]
        if token.value == "STORED":
            storage = ast.GeneratedColumnStorage.STORED
            self.lexer.advance()
        elif token.value == "VIRTUAL":
            storage = ast.GeneratedColumnStorage.VIRTUAL
            self.lexer.advance()
        else:
            storage = None

        return ast.GeneratedColumnConstraint(e, storage=storage)

    @debuggable
    def match_default_clause(self) -> Union[ast.DefaultValue, ast.Node]:
        self.lexer.check(["DEFAULT"])
        token = self.lexer.advance(
            expecting=[
                TokenType.LEFT_PARENTHESIS,
                TokenType.STRING,
                TokenType.INTEGER,
                TokenType.IDENTIFIER,
                "NULL",
                "CURRENT_TIME",
                "CURRENT_DATE",
                "CURRENT_TIMESTAMP",
            ]
        )

        # TODO(2021-05-05): Merge this with match_prefix?
        if token.type == TokenType.LEFT_PARENTHESIS:
            self.lexer.advance()
            e = self.match_expression()
            self.lexer.check([TokenType.RIGHT_PARENTHESIS])
            self.lexer.advance()
            return e
        elif token.type == TokenType.STRING:
            self.lexer.advance()
            return ast.String(token.value)
        elif token.type == TokenType.BLOB:
            self.lexer.advance()
            return ast.Blob(token.value)
        elif token.type == TokenType.INTEGER:
            self.lexer.advance()
            return ast.Integer(int(token.value))
        elif token.type == TokenType.IDENTIFIER:
            if token.value.upper() in ("TRUE", "FALSE"):
                self.lexer.advance()
                return ast.Boolean(token.value.upper() == "TRUE")
            else:
                raise SQLiteParserError
        elif token.type == TokenType.KEYWORD:
            self.lexer.advance()
            if token.value == "NULL":
                return ast.Null()
            elif token.value == "CURRENT_TIME":
                return ast.DefaultValue.CURRENT_TIME
            elif token.value == "CURRENT_TIMESTAMP":
                return ast.DefaultValue.CURRENT_TIMESTAMP
            elif token.value == "CURRENT_DATE":
                return ast.DefaultValue.CURRENT_DATE
            else:
                raise SQLiteParserImpossibleError(token.value)
        else:
            raise SQLiteParserImpossibleError(token.type)

    @debuggable
    def match_on_conflict_clause(self) -> ast.OnConflict:
        self.lexer.check(["ON"])
        self.lexer.advance(expecting=["CONFLICT"])
        strategy = self.lexer.advance(
            expecting=["ROLLBACK", "ABORT", "FAIL", "IGNORE", "REPLACE"]
        ).value
        self.lexer.advance()
        if strategy == "ROLLBACK":
            return ast.OnConflict.ROLLBACK
        elif strategy == "ABORT":
            return ast.OnConflict.ABORT
        elif strategy == "FAIL":
            return ast.OnConflict.FAIL
        elif strategy == "IGNORE":
            return ast.OnConflict.IGNORE
        elif strategy == "REPLACE":
            return ast.OnConflict.REPLACE
        else:
            raise SQLiteParserImpossibleError(strategy)

    @debuggable
    def match_expression(self, precedence: int = -1) -> ast.Expression:
        left = self.match_prefix()

        while True:
            token = self.lexer.current()
            if token is None:
                break

            p = PRECEDENCE.get(token.value)
            if p is None or precedence >= p:
                break

            left = self.match_infix(left, p)
        return left

    @debuggable
    def match_infix(self, left: ast.Expression, precedence: int) -> ast.Infix:
        operator_token = self.lexer.current()
        self.lexer.advance()
        right = self.match_expression(precedence)
        return ast.Infix(operator_token.value, left, right)

    @debuggable
    def match_prefix(self) -> ast.Expression:
        token = self.lexer.current()
        if token.type == TokenType.IDENTIFIER:
            self.lexer.advance()
            return ast.Identifier(token.value)
        elif token.type == TokenType.LEFT_PARENTHESIS:
            self.lexer.advance()
            values = []
            while True:
                e = self.match_expression()
                values.append(e)
                token = self.lexer.check([TokenType.COMMA, TokenType.RIGHT_PARENTHESIS])
                self.lexer.advance()
                if token.type == TokenType.RIGHT_PARENTHESIS:
                    break

            if len(values) == 1:
                return values[0]
            else:
                return ast.ExpressionList(values)
        elif token.type == TokenType.STRING:
            self.lexer.advance()
            return ast.String(token.value)
        elif token.type == TokenType.BLOB:
            self.lexer.advance()
            return ast.Blob(token.value)
        elif token.type == TokenType.INTEGER:
            self.lexer.advance()
            return ast.Integer(int(token.value))
        else:
            raise SQLiteParserError(token.type, token.value)

    def match_identifier_list(self) -> List[str]:
        identifiers = []
        while True:
            token = self.lexer.current()
            if token.type == TokenType.IDENTIFIER:
                identifiers.append(token.value)
                self.lexer.advance()
            elif token.type == TokenType.COMMA:
                self.lexer.advance()
            else:
                break
        return identifiers


# From https://sqlite.org/lang_expr.html
PRECEDENCE = {
    "OR": 0,
    "AND": 1,
    "=": 2,
    "==": 2,
    "!=": 2,
    "<>": 2,
    "IS": 2,
    "IN": 2,
    "LIKE": 2,
    "GLOB": 2,
    "MATCH": 2,
    "REGEXP": 2,
    "<": 3,
    "<=": 3,
    ">": 3,
    ">=": 3,
    "<<": 4,
    ">>": 4,
    "&": 4,
    "|": 4,
    "+": 5,
    "-": 5,
    "*": 6,
    "/": 6,
    "%": 6,
    "||": 7,
}



import os

from setuptools import find_packages, setup

dpath = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dpath, "README.md"), "r") as f:
    long_description = f.read()


setup(
    name="sqliteparser",
    version="0.5.2",
    description="A parser for SQLite's dialect of SQL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Ian Fisher",
    author_email="iafisher@fastmail.com",
    packages=find_packages(exclude=["tests"]),
    install_requires=["attrs >= 20.3.0"],
    project_urls={"Source": "https://github.com/iafisher/sqliteparser"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: SQL",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Database",
    ],
)
